#include <sys/param.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <utmp.h>
#include <signal.h>
#include <curses.h>
#include <stdio.h>
#include <string.h>

#define SAME		0	/* For calls to strcmp */
#define CTRL_G		('G'-'A'+1)
#define CTRL_L		('L'-'A'+1)

/*
	packet.type is one of these --
*/
#define	ERASE		1	/* Erase previous character */
#define	KILL		2	/* Erase line cursor is on */
#define	DISPLAY		3	/* Display character on screen */
#define	ADIOS		4	/* Hang up/disolve connection */
#define	REFRESH		5	/* Redraw screen on CTL-L */
#define	FLASH		6	/* Ring bell/flash screen */

/*
	packet.origin is one of these --
*/
#define	LOCAL		0
#define	REMOTE		1

#define	IPC_ID		230	/* Upper part of key (for uniqueness) */
#define RING_TIME	10	/* Ring cadence in seconds */

int pid = 0;			/* pid for offspring, if non-zero */

int loc_q, rem_q;		/* message queue id's */
key_t my_key, other_key;	/* keys for the two queues */
key_t ftok();			/* convert a file to a key */

char *my_name, *other_name;	/* Both user names */

char *my_tty;			/* Path of my device */
char other_tty[24];		/* Path of other users device */

int erase_ch, kill_ch, eof_ch;	/* Special control chars for user */

WINDOW *topwin, *botwin;	/* curses windows */
WINDOW *topleg, *botleg;	/* windows for legends */
struct utmp *u_elem, *getutent();
struct termio new_ioargs, old_ioargs;

struct packet			/* This is the packet that we ipc */
{
    long type;
    int origin;
    char keypress;
} s_pkt, r_pkt;

void wrapup();			/* Forward reference */
void wrapup_child();		/* Forward reference */
void exit();

main(argc,argv)
    int argc;
    char *argv[];
    {

    if (argc != 2) {
	printf ("usage: phone user\n");
	exit(1);
    }

    signal (SIGINT, wrapup);

    get_names(argv[1]);		/* Get our name and other user's */

    if (!get_ttys())		/* Get our tty and other user's */
	wrapup();

    get_keys();

    if (!open_queues())
	wrapup();

    init_screen();

    if (pid=fork()) {
	scrn_proc();	/* Returns when ADIOS packet received */
	wrapup();
    } else {
	signal (SIGQUIT, wrapup_child);
	key_proc();	/* Never returns, must be killed */
    }
    /* NOTREACHED */
}

void wrapup () {
    int ret_code;			/* Wait insists on an argument */
    struct msqid_ds buf;

    if (pid) {				/* Kill sibbling if present */
	kill (pid, SIGQUIT);
	wait (&ret_code);
    }

    endwin();				/* Clean up curses windows */
    msgctl (loc_q, IPC_RMID, &buf);	/* Remove our message queue */
    send_remote (ADIOS, 0);		/* In case other side is still up */
    exit(0);
}

key_proc() {
    int ch;

    /* Keyboard process */

    init_kbd();

    for (;;) {
	ch=getchar();

        if (ch == erase_ch)
	    send_both (ERASE, 0);

        else if (ch == kill_ch)
	    send_both (KILL, 0);

        else if (ch == CTRL_L)
	    send_local (REFRESH, 0);

        else if (ch == CTRL_G)
	    send_both (FLASH, 0);

	else if (ch == eof_ch)
	    send_local (ADIOS, 0);

        else {
            if (ch == '\r')
		ch = '\n';
	    send_both (DISPLAY, ch);

        }
    }
}

void wrapup_child () {
    reset_kbd();			/* Put ioctl stuff where it was */
    exit(0);
}

scrn_proc() {
    int y, x;
    int ch;
    WINDOW *window;			/* temp curses windows */

    /* Screen process */

    scrn_legend (topleg, other_name, other_tty);
    scrn_legend (botleg, my_name, my_tty);

    wrefresh (topwin);
    wrefresh (botwin);

    do {
	msgrcv (loc_q, &r_pkt, sizeof(r_pkt) - sizeof(r_pkt.type), 0L, 0);

	if (r_pkt.origin == LOCAL)
	    window = botwin;
        else
	    window = topwin;

        ch = r_pkt.keypress;

	switch ((int)r_pkt.type) {

	case ERASE:
	    waddch(window, '\b');
	    wdelch(window);
	    break;

	case KILL:
	    wdeleteln(window);
	    getyx(window, y, x);
	    wmove(window, y, 0);
	    break;

	case DISPLAY:
	    waddch (window, (char)ch);
	    break;

	case ADIOS:
	    if (r_pkt.origin == REMOTE)
	        wprintw(botwin, "\n*** %s has hung up. ***\n", other_name);
	    break;

	case REFRESH:
	    clearok(curscr, TRUE);
	    break;

	case FLASH:
	    flash();
	    break;

	}

	if (r_pkt.origin == REMOTE)
	    wrefresh (topwin);

	wrefresh (botwin);

    } while (r_pkt.type != ADIOS);
}

int open_queues()
    {
    int seconds;

    loc_q = msgget (my_key, 0666 | IPC_CREAT);

    if (loc_q == -1) {
	printf ("Unable to create my queue\n");
	return (FALSE);
    }

    /*
	Stand on our head to test creation of rem_q more often 
	than when we ring the other user.
    */
    for (seconds=RING_TIME; ; seconds++) {
	rem_q = msgget (other_key, 0);

	if (rem_q != -1)
	    break;

	if (seconds == RING_TIME) {
	    seconds = 0;
	    if (!ring (other_tty)) {
		printf ("%s's phone is off the hook (mesg -n).\n\n", other_name);
		return (FALSE);
	    }
	    printf ("Ringing %s on %.14s\n\n", other_name, other_tty);
	}
	if (sleep (1))	/* != 0 if interrupted by other signal */
	    return (FALSE);
    }
    return (TRUE);
}

init_screen() {
    initscr();
    topleg = newwin(1,0,0,0);
    topwin = newwin(10,0,1,0);
    botleg = newwin(1,0,12,0);
    botwin = newwin(10,0,13,0);
    idlok(topwin,TRUE);
    idlok(botwin,TRUE);
    scrollok(topwin,TRUE);
    scrollok(botwin,TRUE);
}

init_kbd() {
    ioctl (0, TCGETA, &old_ioargs);
    erase_ch = old_ioargs.c_cc[VERASE];
    kill_ch = old_ioargs.c_cc[VKILL];
    eof_ch = old_ioargs.c_cc[VEOF];
    new_ioargs = old_ioargs;
    new_ioargs.c_lflag &= ~(ICANON | ECHO);
    new_ioargs.c_cc[VMIN] = (char)1;
    ioctl (0, TCSETA, &new_ioargs);
}

reset_kbd() {
    ioctl (0, TCSETA, &old_ioargs);
}

get_names(cmd_arg)
    char *cmd_arg;
    {
    char *getenv();

    my_name = getenv("LOGNAME");
    other_name = cmd_arg;
    }

int get_ttys()
{
    int found;
    char *ttyname();

    my_tty = ttyname(0);
    found=0;
    while(u_elem=getutent()) {

	if (strcmp(u_elem->ut_user, other_name) == SAME) {
	    found++;
	    strcpy(other_tty, "/dev/");
	    strncpy(&other_tty[5], u_elem->ut_line, sizeof(u_elem->ut_line));
	}
    }
    if (found != 1) {
	if (found == 0)
	    printf("User %s is not logged in.\n", other_name);
	else
	    printf("User %s is logged in more than once.\n", other_name);
	return (FALSE);
    }
    return (TRUE);
}

get_keys()	/* Get unique keys for naming message queues */
{
    my_key = ftok (my_tty, IPC_ID);
    other_key = ftok (other_tty, IPC_ID);
}

int ring (device)
    char *device;
    {
    FILE *out_s;

    if (!(out_s = fopen (device, "w")))
	return (FALSE);

    fprintf (out_s, "\n%s is phoning you.\n", my_name);
    fprintf (out_s, "Type 'phone %s' to answer.\007\n\n", my_name);
    fclose (out_s);
    return (TRUE);
}

scrn_legend (window, user, terminal)
    WINDOW *window;
    char *user;
    char *terminal;
    {
    char buffer[40];

    wattron (window, A_REVERSE | A_BOLD);
    sprintf (buffer, "%s (%s):", user, terminal);
    wprintw (window, "%-*s", COLS, buffer);
    wrefresh (window);
    }

send_local (type, ch)
    int type;
    int ch;
    {
    s_pkt.type = type;
    s_pkt.keypress = ch;
    s_pkt.origin = LOCAL;
    msgsnd (loc_q, &s_pkt, sizeof(s_pkt) - sizeof(s_pkt.type), 0);
    }

send_remote (type, ch)
    int type;
    int ch;
    {
    s_pkt.type = type;
    s_pkt.keypress = ch;
    s_pkt.origin = REMOTE;
    msgsnd (rem_q, &s_pkt, sizeof(s_pkt) - sizeof(s_pkt.type), 0);
    }

send_both (type, ch)
    int type;
    int ch;
    {
    send_local (type, ch);
    send_remote (type, ch);
    }
